#include <stdio.h>

#include "core/utils/utils.h"
#include "core/debug/evemon.h"
#include "core/debug/exception.h"
#include "core/event/notify.h"
#include "core/scope/base.h"
#include "core/scope/ticket.h"
#include "core/scope/access.h"

#include <list>

/* actor class */
class actor : public event::notify::base<>
{
public:
	typedef enum
	{
		EVENT_ACTION=0,
	} event_t;

	actor(void) {}

	virtual void on_notify(int event) override
	{
		_debug("actor::on_notify event=%d", event);
	}

	auto action(void)
	{
		_debug("actor::action");

		return notify(EVENT_ACTION);
	}
};

/* observer class */
class observer : public scope::base
{
public:
	observer(void) {}	

	void init(actor& obj)
	{
		obj.listen(actor::EVENT_ACTION, _SELF(observer, on_action), this);
	}

	void on_action(void)
	{
		_debug("observer::on_action");

		printf("notified using class!\r\n");
	}
};

/* obscure class */
class obscure : public scope::base
{
public:
	class handle : public scope::access<obscure>
	{
	public:
		handle(obscure* parent, const scope::ticket& handle, const std::string& name) : access(parent, handle)
		{
			this->m_name = name;
		}

		void remove(void)
		{
			get()->remove(this->m_name);
		}

		void action(void) const
		{
			get()->action(this->m_name);
		}

	private:
		
		std::string m_name;
	};

public:
	auto add(const std::string& name)
	{
		this->m_list.push_back(name);

		return std::make_unique<handle>(this, get_scope_ticket(), name);
	}

	void print(void) const
	{
		for (auto& v : this->m_list)
			printf("  - %s\r\n", v.c_str());
	}

private:
	void remove(const std::string& name)
	{
		this->m_list.remove(name);
	}

	void action(const std::string& name) const
	{
		_debug("%s performed action!", name.c_str());
	}

private:
	std::list<std::string> m_list;
};

// main program
int main(int argc, char** argv)
{
	try
	{
		// create actor
		actor obj_actor;

		/*
		 *	example 1
		 *	notify class object
		 */
		printf("example 1\r\n");

		{
			{
				// create observer
				observer obj_observer;

				obj_observer.init(obj_actor);

				// trigger actor action
				printf("%zu observers notified\r\n", obj_actor.action());
			}

			// should fail because scope does not exists anymore
			printf("%zu observers notified\r\n", obj_actor.action());
		}

		printf("\r\n");

		/*
		 *	example 2
		 *	notify without class object
		 */
		printf("example 2\r\n");

		{
			{
				scope::ticket ticket;

				obj_actor.listen(actor::EVENT_ACTION, [](void)
					{
						printf("notify using a lambda!\r\n");
					}, ticket);

				// trigger actor action
				printf("%zu observers notified\r\n", obj_actor.action());
			}

			// should fail because scope does not exists anymore
			printf("%zu observers notified\r\n", obj_actor.action());
		}

		printf("\r\n");

		/*
		 *	example 3
		 *	access private parts
		 */
		printf("example 3\r\n");

		{
			{
				// create obscure element
				obscure obj;

				// add some data
				auto handle1 = obj.add("test 1");
				auto handle2 = obj.add("test 2");

				// print
				printf("list:\r\n");
				obj.print();
				printf("\r\n");

				// perform some action
				handle1->action();

				// remove handle1
				handle1->remove();
				printf("remove handle 1:\r\n");
				obj.print();
				printf("\r\n");

				// remove handle2
				handle2->remove();
				printf("remove handle 2:\r\n");
				obj.print();
				printf("\r\n");
			}
		}

		// return successfuly
		return 0;
	}
	catch (exception::base& e)
	{
		printf("error: %s\r\n", e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s\r\n", e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!\r\n");

		_critical("unknown error!");
	}

	// signal error
	return 1;
}
